@@ -1,20 +1,19 @@
 import sys
 import json
-from pathlib import Path
+from pathlib import Path, PosixPath
 import fire
+from typing import Optional, Union
 
-try:
-    from .classes import LogseqPage, LogseqBlock
-except:
-    from classes import LogseqPage, LogseqBlock
+from .classes import LogseqPage, LogseqBlock
+
+__VERSION__ = LogseqPage.__VERSION__
 
-__VERSION__: str = "2.7"
 
 def parse_file(
-    file_path: str = None,
+    file_path: Union[str, PosixPath] = None,
     verbose: bool = False,
-    as_json: bool = False,
-    ):
+    out_format: Optional[str] = None,
+):
     """
     Parameters:
     -----------
@@ -23,8 +22,8 @@ def parse_file(
 
     verbose: bool, default to False
 
-    as_json: default to False.
-        If True will output a json string meant to be piped to jq.
+    out_format: default to None
+        Either 'json' or 'toml'. For example can be piped directly to jq
     """
     if file_path is not None:
         assert Path(file_path).exists(), f"{file_path} not found"
@@ -34,21 +33,19 @@ def parse_file(
         content = sys.stdin.read()
 
     parsed = LogseqPage(
-            content=content,
-            verbose=verbose,
-            )
+        content=content,
+        verbose=verbose,
+    )
 
-    if as_json:
-        return json.dumps(
-            parsed.as_json(),
-            indent=4,
-            ensure_ascii=False
-        )
+    if out_format:
+        return parsed.export(format=out_format)
     else:
         return parsed
 
+
 def cli() -> None:
     fire.Fire(parse_file)
 
+
 if __name__ == "__main__":
     fire.Fire(parse_file)
