@@ -1,7 +1,27 @@
+import sys
 import textwrap
-from pathlib import Path
+from typing import Union
+from pathlib import Path, PosixPath
 import uuid
 import re
+import json
+import rtoml as toml
+
+__VERSION__: str = "2.13"
+
+# only use beartype if its installed
+try:
+    from beartype import beartype
+except Exception as err:
+    pass
+
+# if used in a tqdm loop, it's annoying to have the prints appear
+# if tqdm is found, use it instead
+try:
+    from tqdm import tqdm
+    print = lambda x: tqdm.write(str(x))
+except Exception as err:
+    pass
 
 
 class LogseqPage:
@@ -12,6 +32,8 @@ class LogseqPage:
             list of LogseqBlock objects
         * page_properties
             can be edited like a normal dict, as opposed to the block properties
+        * __VERSION__
+            version of the LogseqMarkdownParser
 
     Methods:
         * export_to
@@ -20,19 +42,20 @@ class LogseqPage:
             page_property attribute
 
     """
+    __VERSION__ = __VERSION__
     PAGE_PROP_REGEX = re.compile(r"(\w[\w_-]*\w:: .+)")
+
     def __init__(
         self,
         content: str,
         check_parsing: bool = False,
         verbose: bool = False,
-        ) -> None:
+    ) -> None:
         self.verbose = verbose
         assert isinstance(content, str), (
             f"content must be of type string, not '{type(content)}'")
 
         content = content.strip()
-        assert content, "Cannot supply empty content"
 
         # detect each block (read each line then merge with the latest block)
         lines = content.split("\n")
@@ -43,7 +66,8 @@ class LogseqPage:
             for li in lines
         ]
 
-        assert lines[0].lstrip().startswith("-") or ":: " in lines[0] or (len(lines)==1 and not lines[0].strip()), f"First line of document must start with '[ \t]*- ' or contain a page property or the document must be empty"
+        assert lines[0].lstrip().startswith("-") or lines[0].lstrip().startswith("#") or ":: " in lines[0] or (len(lines) == 1 and not lines[0].strip()
+                                                                          ), f"First line of document must start with '[ \t]*- ' or '[ \t]*#' or contain a page property or the document must be empty"
         lines = [l for l in lines if l.strip()]  # remove empty lines
         pageprop = ""  # as string first
         first_block_reached = False
@@ -89,11 +113,12 @@ class LogseqPage:
         # parse each block
         self.blocks = []
         for index, block_str in enumerate(blocks):
-            assert isinstance(block_str, str), f"block is not string: '{block_str}'"
+            assert isinstance(
+                block_str, str), f"block is not string: '{block_str}'"
             block = LogseqBlock(
-                    content=block_str,
-                    verbose=self.verbose,
-                    )
+                content=block_str,
+                verbose=self.verbose,
+            )
             assert block.content == block_str.replace(u"\xa0", u" "), (
                 "block content modifying unexpectedly")
 
@@ -115,7 +140,8 @@ class LogseqPage:
         if reformed.replace(u"\xa0", u" ") != content.replace(u"\xa0", u" "):
             print("\n\nError: file content differed after parsing:")
             print(f"Length: '{len(reformed)}' vs '{len(content)}'")
-            print("Nb lines: '" + str(len(reformed.split('\n'))) + " vs '" + str(len(content.split('\n'))) + "'")
+            print("Nb lines: '" + str(len(reformed.split('\n'))) +
+                  " vs '" + str(len(content.split('\n'))) + "'")
             spco = content.split("\n")
             spref = reformed.split("\n")
             nref = len(spref)
@@ -140,7 +166,8 @@ class LogseqPage:
     @property
     def content(self) -> str:
         "return the concatenated list of each block of the page"
-        temp = "\n".join([f"{k}:: {v}" for k, v in self.page_properties.items()])
+        temp = "\n".join(
+            [f"{k}:: {v}" for k, v in self.page_properties.items()])
         if self.blocks:
             if temp:
                 temp += "\n"
@@ -184,10 +211,10 @@ class LogseqPage:
 
     def export_to(
         self,
-        file_path: str,
+        file_path: Union[str, PosixPath],
         overwrite: bool = False,
         allow_empty: bool = False,
-        ) -> None:
+    ) -> None:
         """
         export the blocks to file_path
         """
@@ -212,21 +239,32 @@ class LogseqPage:
     def __repr__(self) -> str:
         return f"LogseqPage({self.__str__()})"
 
-    def as_json(self) -> list[dict]:
-        """returns a json format string for the whole text. Meant to be piped
-        to jq"""
-        return [block.as_json() for block in self.blocks]
+    def export(self, format: str) -> Union[list[dict], str]:
+        """returns the whole logseq page formatted.
+        Expected formats are "list_of_dict", "json", "toml".
+        'json' for example can be piped directly to jq in a shell.
+        """
+        cont = [block.export(format="dict") for block in self.blocks]
+        if format == "list_of_dict":
+            return cont
+        elif format == "json":
+            return json.dumps(cont, ensure_ascii=False, indent=2)
+        elif format == "toml":
+            return toml.dumps(cont, pretty=True)
+        else:
+            raise ValueError(format)
 
 
 class LogseqBlock:
     BLOCK_PROP_REGEX = re.compile(r"[ \t]+(\w[\w_-]*\w:: .+)")
     INDENT_REGEX = re.compile(r"^[ \t]*")
+    __VERSION__ = __VERSION__
 
     def __init__(
             self,
             content: str,
             verbose: bool = False,
-            ) -> None:
+    ) -> None:
         """
         Class with the following new attributes:
             indentation_level: in number of spaces, with tab=4
@@ -260,8 +298,8 @@ class LogseqBlock:
         self.verbose = verbose
         content = content.replace(u'\xa0', u' ')  # replace by regular space
         self._blockvalues = {
-                'content': content,
-            }
+            'content': content,
+        }
         if "id" in self.properties:
             self._blockvalues["UUID"] = self.properties["id"]
         else:
@@ -286,7 +324,8 @@ class LogseqBlock:
     def content(self, new: str) -> str:
         old = self._blockvalues["content"]
         assert isinstance(new, str), "new content must be a string"
-        assert new.lstrip().startswith("-") or ":: " in new, "stripped new content must start with '-' or be a property"
+        assert new.lstrip().startswith(
+            "-") or ":: " in new, "stripped new content must start with '-' or be a property"
         new = new.replace(u'\xa0', u' ')
         if new != old:
             self._changed = True
@@ -310,7 +349,7 @@ class LogseqBlock:
             "block intentation level apparently failed to be set")
 
     @property
-    def TODO_state(self) -> str:
+    def TODO_state(self) -> Union[None, str]:
         return self._get_TODO_state()
 
     @TODO_state.setter
@@ -328,15 +367,15 @@ class LogseqBlock:
         if old != new:
             if new is None:  # just delete
                 self.content = self.content.replace(
-                        f"- {old} ",
-                        "- ",
-                        1,
-                        )
+                    f"- {old} ",
+                    "- ",
+                    1,
+                )
             else:
                 self.content = self.content.replace(
-                        f"- {old} ",
-                        f"- {new} ",
-                        1)
+                    f"- {old} ",
+                    f"- {new} ",
+                    1)
             self._changed = True
 
     @property
@@ -345,13 +384,15 @@ class LogseqBlock:
 
     @properties.setter
     def property_failedsetter(self, *args, **kwargs) -> None:
-        raise Exception("To modify the properties you must use self.set_property(key, value)")
+        raise Exception(
+            "To modify the properties you must use self.set_property(key, value)")
 
     def _get_properties(self) -> dict:
         prop = re.findall(self.BLOCK_PROP_REGEX, self.content)
         properties = {}
         for found in prop:
-            assert found == found.lstrip(), f"REGEX match an incorrect property: {found}"
+            assert found == found.lstrip(
+            ), f"REGEX match an incorrect property: {found}"
 
             try:
                 key, value = found.split(":: ")
@@ -385,10 +426,12 @@ class LogseqBlock:
             try:
                 value = str(value)
             except Exception as err:
-                raise Exception(f"Failed to parse as string: '{value}' (err:{err})")
+                raise Exception(
+                    f"Failed to parse as string: '{value}' (err:{err})")
 
             assert value, f"Cannot add empty property"
-            assert len(value.splitlines()) == 1, "cannot add property containing newlines"
+            assert len(value.splitlines()
+                       ) == 1, "cannot add property containing newlines"
 
             old_content = self.content
             if key in self.properties:  # edit prop
@@ -396,9 +439,9 @@ class LogseqBlock:
                 assert self.content.count(f"  {key}:: {old_val}") == 1, (
                     "unable to find key/val pair: {key}/{old_val}")
                 self.content = self.content.replace(
-                        f"  {key}:: {old_val}",
-                        f"  {key}:: {value}",
-                        )
+                    f"  {key}:: {old_val}",
+                    f"  {key}:: {value}",
+                )
             else:  # add prop
                 new = "\n" + "\t" * (self.indentation_level // 4)
                 new += f"  {key}:: {value}"
@@ -417,31 +460,40 @@ class LogseqBlock:
         else:
             assert key in self.properties, "key not found in properties"
             assert self.content.count(f"  {key}:: ") == 1, (
-                    f"invalid number of key {key} in {self.content}")
+                f"invalid number of key {key} in {self.content}")
             temp = []
             for line in self.content.split("\n"):
                 if not re.search(rf"[ \t]+{key}:: ", line):
                     temp.append(line)
             new_content = "\n".join(temp)
             assert new_content.count(f"{key}::") == 0, (
-                    f"invalid number of key {key} in {new_content}")
+                f"invalid number of key {key} in {new_content}")
             self.content = new_content
             self._changed = True
 
             assert key not in self.properties, (
                 "key apparently failed to be deleted")
 
-    def as_json(self) -> dict:
-        """returns the block as json representation."""
-        return {
-                "block_properties": self.properties,
-                "block_content": self.content,
-                "block_indentation_level": self.indentation_level,
-                "block_TODO_state": self.TODO_state,
-                "block_UUID": self.UUID,
-                }
+    def export(self, format: str) -> Union[dict, str]:
+        """export the block. Formats are 'dict', 'json', 'toml'"""
+        assert format in ["dict", "json", "toml"], "support export format are dict, json, toml"
+        d = {
+            "block_properties": self.properties,
+            "block_content": self.content,
+            "block_indentation_level": self.indentation_level,
+            "block_TODO_state": self.TODO_state,
+            "block_UUID": self.UUID,
+        }
+        if format == "dict":
+            return d
+        elif format == "json":
+            return json.dumps(d, ensure_ascii=False, indent=2)
+        elif format == "toml":
+            return toml.dumps(d, pretty=True)
+        else:
+            raise ValueError(format)
 
-    def _get_TODO_state(self) -> str:
+    def _get_TODO_state(self) -> Union[None, str]:
         TODO_state = None
         for keyword in ["TODO", "DOING", "NOW", "LATER", "DONE"]:
             if re.search(f"- {keyword} .*", self.content.lstrip()):
@@ -454,11 +506,11 @@ class LogseqBlock:
     def _get_indentation(self) -> int:
         """count the leading spaces of a block to know the indentation level"""
         n = len(
-                re.search(
-                    self.INDENT_REGEX,
-                    self.content.replace("\t", " " * 4)
-                    ).group(0)
-                )
+            re.search(
+                self.INDENT_REGEX,
+                self.content.replace("\t", " " * 4)
+            ).group(0)
+        )
         return n
 
     @property
@@ -469,9 +521,15 @@ class LogseqBlock:
         return self._blockvalues["UUID"]
 
     @UUID.setter
-    def UUID(self, new):
+    def UUID(self, new: str) -> None:
         assert isinstance(new, str), "new id must be a string"
         assert new.count("-") == 4, "new id must contain 4 -"
-        assert new.replace("-", "").isalnum(), "new id does not look like a UUID4"
+        assert new.replace(
+            "-", "").isalnum(), "new id does not look like a UUID4"
         self.set_property(key="id", value=new)
         self._changed = True
+
+
+if "beartype" in sys.modules:
+    LogseqBlock = beartype(LogseqBlock)
+    LogseqPage = beartype(LogseqPage)
